/*
 * This file is part of the MicroPython project, http://micropython.org/
 *
 * The MIT License (MIT)
 *
 * Copyright (c) 2023 Mariano Goluboff
 * Copyright (c) 2017 Linaro Limited
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

 #include "py/mpconfig.h"
 #if (CONFIG_NET_SOCKETS && !CONFIG_EXCLUDE_PY_SOCKETS)
 
 #include "py/runtime.h"
 #include "py/stream.h"
 #include "mpthreadport.c"
 
 #include <stdio.h>
 #include <errno.h>
 #include <zephyr/kernel.h>
 // Zephyr's generated version header
 #include <version.h>
 #include <zephyr/net/net_pkt.h>
 #include <zephyr/net/dns_resolve.h>
 #include <zephyr/net/socket.h>
 #include <zephyr/posix/fcntl.h>

 #include <zephyr/net/tls_credentials.h>
 #if (CONFIG_NET_SOCKETS_SOCKOPT_TLS || CONFIG_NET_SOCKETS_OFFLOAD)
 #endif
 
 #define DEBUG_PRINT 1
 #if DEBUG_PRINT // print debugging info
 #define DEBUG_printf printf
 #else // don't print debugging info
 #define DEBUG_printf(...) (void)0
 #endif

#include <string.h>
#include <zephyr/kernel.h>
#include <stdlib.h>
#include <zephyr/net/socket.h>
#include <zephyr/net/conn_mgr_monitor.h>
#include <zephyr/net/conn_mgr_connectivity.h>
#include <zephyr/net/tls_credentials.h>

#if defined(CONFIG_POSIX_API)
#include <zephyr/posix/arpa/inet.h>
#include <zephyr/posix/netdb.h>
#include <zephyr/posix/unistd.h>
#include <zephyr/posix/sys/socket.h>
#endif





// Potentiallt necessary...?
// #include "psa/storage_common.h"
// #include "psa/protected_storage.h"

 #define GOOGLE_CA_SEC_TAG 16842753  // any unused sec_tag you pick

/* Binary DER-encoded certificate data */
#include "cert/r1.crt.inc"
#define MBEDTLS_SSL_TLS_C 1
 typedef struct _socket_obj_t {
     mp_obj_base_t base;
     int ctx;
 
     #define STATE_NEW 0
     #define STATE_CONNECTING 1
     #define STATE_CONNECTED 2
     #define STATE_PEER_CLOSED 3
     int8_t state;
     sa_family_t family;
 } socket_obj_t;
 
//  static struct dns_resolve_context my_ctx;
 STATIC const mp_obj_type_t socket_type;
 
 // Helper functions
 
 #define RAISE_ERRNO(x) { int _err = x; if (_err < 0) mp_raise_OSError(-_err); }
 #define RAISE_SOCK_ERRNO(x) { if ((int)(x) == -1) mp_raise_OSError(errno); }
 
 STATIC void socket_check_closed(socket_obj_t *socket) {
     if (socket->ctx == -1) {
         // already closed
         mp_raise_OSError(EBADF);
     }
 }
 
 STATIC void parse_inet_addr(socket_obj_t *socket, mp_obj_t addr_in, struct sockaddr *sockaddr) {
     // We employ the fact that port and address offsets are the same for IPv4 & IPv6
     struct sockaddr_in *sockaddr_in = (struct sockaddr_in *)sockaddr;
 
     mp_obj_t *addr_items;
 
     sockaddr_in->sin_family = socket->family;
     size_t address_array_size = (sockaddr_in->sin_family == AF_INET6) ? 4 : 2;
     mp_obj_get_array_fixed_n(addr_in, address_array_size, &addr_items);
     RAISE_ERRNO(net_addr_pton(sockaddr_in->sin_family, mp_obj_str_get_str(addr_items[0]), &sockaddr_in->sin_addr));
     sockaddr_in->sin_port = htons(mp_obj_get_int(addr_items[1]));
 }
 
 STATIC mp_obj_t format_inet_addr(struct sockaddr *addr, mp_obj_t port) {
     // We employ the fact that port and address offsets are the same for IPv4 & IPv6
     struct sockaddr_in6 *sockaddr_in6 = (struct sockaddr_in6 *)addr;
     char buf[40];
     net_addr_ntop(addr->sa_family, &sockaddr_in6->sin6_addr, buf, sizeof(buf));
     mp_obj_tuple_t *tuple = mp_obj_new_tuple(addr->sa_family == AF_INET ? 2 : 4, NULL);
 
     tuple->items[0] = mp_obj_new_str(buf, strlen(buf));
     // We employ the fact that port offset is the same for IPv4 & IPv6
     // not filled in
     // tuple->items[1] = mp_obj_new_int(ntohs(((struct sockaddr_in*)addr)->sin_port));
     tuple->items[1] = port;
 
     if (addr->sa_family == AF_INET6) {
         tuple->items[2] = MP_OBJ_NEW_SMALL_INT(0); // flow_info
         tuple->items[3] = MP_OBJ_NEW_SMALL_INT(sockaddr_in6->sin6_scope_id);
     }
 
     return MP_OBJ_FROM_PTR(tuple);
 }
 
 socket_obj_t *socket_new(void) {
     socket_obj_t *socket = m_new_obj(socket_obj_t);
     socket->base.type = (mp_obj_t)&socket_type;
     socket->state = STATE_NEW;
     return socket;
 }
 
 // Methods
 
 STATIC void socket_print(const mp_print_t *print, mp_obj_t self_in, mp_print_kind_t kind) {
     socket_obj_t *self = self_in;
     if (self->ctx == -1) {
         mp_printf(print, "<socket NULL>");
     } else {
         mp_printf(print, "<socket %p family=%d>", self->ctx, self->family);
     }
 }
 #include <time.h>
 #include <date_time.h>
 #include <stdio.h> // For printk

void set_nordic_date_time(void) {
    struct tm dt;
    int ret;

    // Set the desired date and time
    // Example: May 21, 2024, 10:30:00 UTC
    dt.tm_year = 125;
    dt.tm_mon = 5;
    dt.tm_mday = 21;
    dt.tm_hour = 11; // Use UTC hours
    dt.tm_min = 25;
    dt.tm_sec = 0;
    dt.tm_wday = 2; // 0=Sunday, 1=Monday, ..., 6=Saturday (Tuesday)
    // dt.tz = 0; // Timezone offset in minutes from UTC (0 for UTC)

    printk("Attempting to set date and time (YYYY-MM-DD HH:MM:SS) to %04d-%02d-%02d %02d:%02d:%02d...\n",
           dt.tm_year, dt.tm_mon, dt.tm_mday, dt.tm_hour, dt.tm_min, dt.tm_sec);

    // Initialize the date_time library if not already done
    // Depending on your application's init, this might be needed.
    // Check date_time_init() API if required. For simplicity, we assume
    // it might be handled elsewhere or implicitly.

    ret = date_time_set(&dt);

    if (ret == 0) {
        printk("Date and time set successfully using date_time_set.\n");
    } else {
        printk("Failed to set date and time using date_time_set, error: %d\n", ret);
    }
}
#include <mbedtls/debug.h>
 STATIC mp_obj_t socket_make_new(const mp_obj_type_t *type, size_t n_args, size_t n_kw, const mp_obj_t *args) {
     mp_arg_check_num(n_args, n_kw, 0, 4, false);
 
    //  set_nordic_date_time();
     mbedtls_debug_set_threshold(4); 
     socket_obj_t *socket = socket_new();
 
     int family = AF_INET;
     int socktype = SOCK_STREAM;
     int proto = -1;
 
     if (n_args >= 1) {
         family = mp_obj_get_int(args[0]);
         if (n_args >= 2) {
             socktype = mp_obj_get_int(args[1]);
             if (n_args >= 3) {
                 proto = mp_obj_get_int(args[2]);
             }
         }
     }
 
     if (proto == -1) {
         proto = IPPROTO_TLS_1_2;
         if (socktype != SOCK_STREAM) {
             proto = IPPROTO_UDP;
         }
     }
     
    gc_collect();

     int res = tls_credential_add(GOOGLE_CA_SEC_TAG, TLS_CREDENTIAL_CA_CERTIFICATE, r1_crt, r1_crt_len);
     RAISE_SOCK_ERRNO(res);
     printk("Trying to initialize socket...\n");
     printk("Family: %d, socktype: %d, proto: %d \n", family, socktype, proto);
     socket->ctx = zsock_socket(family, socktype, proto);
     printk("Returned value: %d \n", socket->ctx);
     RAISE_SOCK_ERRNO(socket->ctx);
     printk("Done initializing socket! \n");
     socket->family = family;
     return MP_OBJ_FROM_PTR(socket);
 }  
 
//  #if CONFIG_NET_SOCKETS_OFFLOAD
 STATIC mp_obj_t socket_tlswrap(size_t n_args, const mp_obj_t *pos_args, mp_map_t *kwargs) {
     enum { ARG_sec_tag, ARG_verify, ARG_hostname };
     
     static const mp_arg_t allowed_args[] = {
         { MP_QSTR_sec_tag, MP_ARG_REQUIRED | MP_ARG_INT },
         { MP_QSTR_verify, MP_ARG_INT, {.u_int = TLS_PEER_VERIFY_REQUIRED} },
         { MP_QSTR_hostname, MP_ARG_KW_ONLY | MP_ARG_OBJ, {.u_obj = mp_const_none} },
     };
 
     // parse args
     mp_arg_val_t args[MP_ARRAY_SIZE(allowed_args)];
     mp_arg_parse_all(n_args - 1, pos_args + 1, kwargs, MP_ARRAY_SIZE(allowed_args), allowed_args, args);

     socket_obj_t *socket = pos_args[0];
     sec_tag_t sec_tag = args[ARG_sec_tag].u_int;
     int res;
     printk("Trying 1...\n");
     int verify = args[ARG_verify].u_int;
     res = setsockopt(socket->ctx, SOL_TLS, TLS_PEER_VERIFY, &verify, sizeof(verify));
    //  printk("Failed to setup TLS sec tag, err %d\n", errno);
     RAISE_SOCK_ERRNO(res);
     printk("1 is done\n");
     res = setsockopt(socket->ctx, SOL_TLS, TLS_SEC_TAG_LIST, &sec_tag, sizeof(sec_tag));
     RAISE_SOCK_ERRNO(res);
     printk("2 is done\n");
     if (args[ARG_hostname].u_obj != mp_const_none) {
         const char *hostname = mp_obj_str_get_str(args[ARG_hostname].u_obj);
         setsockopt(socket->ctx, SOL_TLS, TLS_HOSTNAME, hostname, strlen(hostname));
         RAISE_SOCK_ERRNO(res);
         printk("3 is done\n");
     }
     return mp_const_none;
 }
 MP_DEFINE_CONST_FUN_OBJ_KW(socket_tlswrap_obj, 1, socket_tlswrap);
//  #endif

 STATIC mp_obj_t socket_bind(mp_obj_t self_in, mp_obj_t addr_in) {
     socket_obj_t *socket = self_in;
     socket_check_closed(socket);
 
     struct sockaddr sockaddr;
     parse_inet_addr(socket, addr_in, &sockaddr);
 
     int res = zsock_bind(socket->ctx, &sockaddr, sizeof(sockaddr));
     RAISE_SOCK_ERRNO(res);
 
     return mp_const_none;
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_2(socket_bind_obj, socket_bind);
 
 STATIC mp_obj_t socket_connect(mp_obj_t self_in, mp_obj_t addr_in) {
     socket_obj_t *socket = self_in;
     socket_check_closed(socket);
 
     struct sockaddr sockaddr;
     parse_inet_addr(socket, addr_in, &sockaddr);
 
     int res = zsock_connect(socket->ctx, &sockaddr, sizeof(sockaddr));
     RAISE_SOCK_ERRNO(res);
 
     return mp_const_none;
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_2(socket_connect_obj, socket_connect);
 
 // method socket.listen([backlog])
 STATIC mp_obj_t socket_listen(size_t n_args, const mp_obj_t *args) {
     socket_obj_t *socket = args[0];
     socket_check_closed(socket);
 
     mp_int_t backlog = MICROPY_PY_SOCKET_LISTEN_BACKLOG_DEFAULT;
     if (n_args > 1) {
         backlog = mp_obj_get_int(args[1]);
         backlog = (backlog < 0) ? 0 : backlog;
     }
 
     int res = zsock_listen(socket->ctx, backlog);
     RAISE_SOCK_ERRNO(res);
 
     return mp_const_none;
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(socket_listen_obj, 1, 2, socket_listen);
 
 STATIC mp_obj_t socket_accept(mp_obj_t self_in) {
     socket_obj_t *socket = self_in;
     socket_check_closed(socket);
 
     struct sockaddr sockaddr;
     socklen_t addrlen = sizeof(sockaddr);
     int ctx = zsock_accept(socket->ctx, &sockaddr, &addrlen);
 
     socket_obj_t *socket2 = socket_new();
     socket2->ctx = ctx;
 
     mp_obj_tuple_t *client = mp_obj_new_tuple(2, NULL);
     client->items[0] = MP_OBJ_FROM_PTR(socket2);
     // TODO
     client->items[1] = mp_const_none;
 
     return MP_OBJ_FROM_PTR(client);
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_1(socket_accept_obj, socket_accept);
 
 STATIC mp_uint_t sock_write(mp_obj_t self_in, const void *buf, mp_uint_t size, int *errcode) {
     socket_obj_t *socket = self_in;
     if (socket->ctx == -1) {
         // already closed
         *errcode = EBADF;
         return MP_STREAM_ERROR;
     }
 
     ssize_t len = zsock_send(socket->ctx, buf, size, 0);
     if (len == -1) {
         *errcode = errno;
         return MP_STREAM_ERROR;
     }
 
     return len;
 }
 
 STATIC mp_obj_t socket_send(mp_obj_t self_in, mp_obj_t buf_in) {
     mp_buffer_info_t bufinfo;
     mp_get_buffer_raise(buf_in, &bufinfo, MP_BUFFER_READ);
     int err = 0;
     mp_uint_t len = sock_write(self_in, bufinfo.buf, bufinfo.len, &err);
     if (len == MP_STREAM_ERROR) {
         mp_raise_OSError(err);
     }
     return mp_obj_new_int_from_uint(len);
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_2(socket_send_obj, socket_send);
 
 STATIC mp_uint_t sock_read(mp_obj_t self_in, void *buf, mp_uint_t max_len, int *errcode) {
     socket_obj_t *socket = self_in;
     if (socket->ctx == -1) {
         // already closed
         *errcode = EBADF;
         return MP_STREAM_ERROR;
     }
 
     ssize_t recv_len = zsock_recv(socket->ctx, buf, max_len, 0);
     if (recv_len == -1) {
         *errcode = errno;
         return MP_STREAM_ERROR;
     }
 
     return recv_len;
 }
 
 STATIC mp_obj_t socket_recv(mp_obj_t self_in, mp_obj_t len_in) {
     mp_int_t max_len = mp_obj_get_int(len_in);
     vstr_t vstr;
     // +1 to accommodate for trailing \0
     vstr_init_len(&vstr, max_len + 1);
 
     int err;
     mp_uint_t len = sock_read(self_in, vstr.buf, max_len, &err);
 
     if (len == MP_STREAM_ERROR) {
         vstr_clear(&vstr);
         mp_raise_OSError(err);
     }
 
     if (len == 0) {
         vstr_clear(&vstr);
         return mp_const_empty_bytes;
     }
 
     vstr.len = len;
     return mp_obj_new_bytes_from_vstr(&vstr);
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_2(socket_recv_obj, socket_recv);
 
 STATIC mp_obj_t socket_setsockopt(size_t n_args, const mp_obj_t *args) {
     (void)n_args; // always 4
     mp_warning(MP_WARN_CAT(RuntimeWarning), "setsockopt() not implemented");
     return mp_const_none;
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(socket_setsockopt_obj, 4, 4, socket_setsockopt);

 // method socket.settimeout(value)
// timeout=0 means non-blocking
// timeout=None means blocking
// otherwise, timeout is in seconds
// static mp_obj_t socket_settimeout(mp_obj_t self_in, mp_obj_t timeout_in) {
//     mod_network_socket_obj_t *self = MP_OBJ_TO_PTR(self_in);
//     mp_uint_t timeout;
//     if (timeout_in == mp_const_none) {
//         timeout = -1;
//     } else {
//         #if MICROPY_PY_BUILTINS_FLOAT
//         timeout = (mp_uint_t)(MICROPY_FLOAT_CONST(1000.0) * mp_obj_get_float(timeout_in));
//         #else
//         timeout = 1000 * mp_obj_get_int(timeout_in);
//         #endif
//     }
//     if (self->nic == MP_OBJ_NULL) {
//         #if MICROPY_PY_SOCKET_EXTENDED_STATE
//         // store the timeout in the socket state until a NIC is bound
//         self->timeout = timeout;
//         #else
//         // not connected
//         mp_raise_OSError(MP_ENOTCONN);
//         #endif
//     } else {
//         int _errno;
//         if (self->nic_protocol->settimeout(self, timeout, &_errno) != 0) {
//             mp_raise_OSError(_errno);
//         }
//     }
//     return mp_const_none;
// }
// static MP_DEFINE_CONST_FUN_OBJ_2(socket_settimeout_obj, socket_settimeout);
 
 STATIC mp_obj_t socket_setblocking(mp_obj_t self_in, mp_obj_t blocking) {
     socket_obj_t *socket = self_in;
     int ret = zsock_fcntl(socket->ctx, F_GETFL, 0);
     if (ret == -1) {
         mp_raise_OSError(-errno);
     }
     if (mp_obj_is_true(blocking)) {
         ret = zsock_fcntl(socket->ctx, F_SETFL, ret & ~(O_NONBLOCK));
     } else {
         ret = zsock_fcntl(socket->ctx, F_SETFL, ret | O_NONBLOCK);
     }
     if (ret == -1) {
         mp_raise_OSError(-errno);
     }
     return mp_const_none;
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_2(socket_setblocking_obj, socket_setblocking);
 
 STATIC mp_obj_t socket_makefile(size_t n_args, const mp_obj_t *args) {
     (void)n_args;
     return args[0];
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(socket_makefile_obj, 1, 3, socket_makefile);
 
 STATIC mp_uint_t sock_ioctl(mp_obj_t o_in, mp_uint_t request, uintptr_t arg, int *errcode) {
     socket_obj_t *socket = o_in;
     (void)arg;
     switch (request) {
         case MP_STREAM_CLOSE:
             if (socket->ctx != -1) {
                 int res = zsock_close(socket->ctx);
                 RAISE_SOCK_ERRNO(res);
                 if (res == -1) {
                     *errcode = errno;
                     return MP_STREAM_ERROR;
                 }
                 socket->ctx = -1;
             }
             return 0;
 
         default:
             *errcode = MP_EINVAL;
             return MP_STREAM_ERROR;
     }
 }
 
 #if CONFIG_PDN
 STATIC mp_obj_t socket_pdn(mp_obj_t self_in, mp_obj_t id) {
     if (!mp_obj_is_int(id)) {
         mp_raise_TypeError(MP_ERROR_TEXT("ID must be an integer"));
     }
     socket_obj_t *socket = self_in;
     int pdn_id = mp_obj_get_int(id);
     int ret = setsockopt(socket->ctx, SOL_SOCKET, SO_BINDTOPDN, &pdn_id, sizeof(pdn_id));
     if (ret) {
         mp_raise_OSError(ret);
     }
     return mp_const_none;
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_2(socket_pdn_obj, socket_pdn);
 #endif
 
 STATIC const mp_rom_map_elem_t socket_locals_dict_table[] = {
     { MP_ROM_QSTR(MP_QSTR___del__), MP_ROM_PTR(&mp_stream_close_obj) },
     { MP_ROM_QSTR(MP_QSTR_close), MP_ROM_PTR(&mp_stream_close_obj) },
     { MP_ROM_QSTR(MP_QSTR_bind), MP_ROM_PTR(&socket_bind_obj) },
     { MP_ROM_QSTR(MP_QSTR_connect), MP_ROM_PTR(&socket_connect_obj) },
     { MP_ROM_QSTR(MP_QSTR_listen), MP_ROM_PTR(&socket_listen_obj) },
     { MP_ROM_QSTR(MP_QSTR_accept), MP_ROM_PTR(&socket_accept_obj) },
     { MP_ROM_QSTR(MP_QSTR_send), MP_ROM_PTR(&socket_send_obj) },
     { MP_ROM_QSTR(MP_QSTR_recv), MP_ROM_PTR(&socket_recv_obj) },
     { MP_ROM_QSTR(MP_QSTR_setsockopt), MP_ROM_PTR(&socket_setsockopt_obj) },
    //  { MP_ROM_QSTR(MP_QSTR_settimeout), MP_ROM_PTR(&socket_settimeout_obj) },
     { MP_ROM_QSTR(MP_QSTR_setblocking), MP_ROM_PTR(&socket_setblocking_obj) },
 
     { MP_ROM_QSTR(MP_QSTR_read), MP_ROM_PTR(&mp_stream_read_obj) },
     { MP_ROM_QSTR(MP_QSTR_readinto), MP_ROM_PTR(&mp_stream_readinto_obj) },
     { MP_ROM_QSTR(MP_QSTR_readline), MP_ROM_PTR(&mp_stream_unbuffered_readline_obj) },
     { MP_ROM_QSTR(MP_QSTR_write), MP_ROM_PTR(&mp_stream_write_obj) },
     { MP_ROM_QSTR(MP_QSTR_makefile), MP_ROM_PTR(&socket_makefile_obj) },
    //  #if CONFIG_NET_SOCKETS_OFFLOAD
     { MP_ROM_QSTR(MP_QSTR_tlswrap), MP_ROM_PTR(&socket_tlswrap_obj) },
     #if CONFIG_PDN
     { MP_ROM_QSTR(MP_QSTR_pdn), MP_ROM_PTR(&socket_pdn_obj) },
     #endif
    //  #endif
 };
 STATIC MP_DEFINE_CONST_DICT(socket_locals_dict, socket_locals_dict_table);
 
 STATIC const mp_stream_p_t socket_stream_p = {
     .read = sock_read,
     .write = sock_write,
     .ioctl = sock_ioctl,
 };
 
 STATIC MP_DEFINE_CONST_OBJ_TYPE(
     socket_type,
     MP_QSTR_socket,
     MP_TYPE_FLAG_NONE,
     make_new, socket_make_new,
     print, socket_print,
     protocol, &socket_stream_p,
     locals_dict, &socket_locals_dict
     );
 
 //
 // getaddrinfo() implementation
 //
 
 typedef struct _getaddrinfo_state_t {
     mp_obj_t result;
     struct k_sem sem;
     mp_obj_t port;
     int status;
     mp_thread_t *calling_thread_state;
 } getaddrinfo_state_t;
 
 void dns_resolve_cb(enum dns_resolve_status status, struct dns_addrinfo *info, void *user_data) {
    if (!user_data) {
        printk("DNS CALLBACK: user_data is NULL!\n");
        return;
    }

    getaddrinfo_state_t *state = user_data;

    // Set the MicroPython thread state for this callback
    mp_thread_set_state(state->calling_thread_state);

    printk("DNS CALLBACK: status=%d\n", status);

    // Process valid info
    if (info != NULL && info->ai_addr.sa_family != 0) {
        // Safety: make sure address is valid
        if (info->ai_addr.sa_family != AF_INET && info->ai_addr.sa_family != AF_INET6) {
            printk("DNS CALLBACK: invalid sa_family=%d\n", info->ai_addr.sa_family);
            // We could set an error here, but for now, just ignore this invalid entry.
            // state->status = -EINVAL; // Consider setting this if needed
            return;
        }

        mp_obj_tuple_t *tuple = mp_obj_new_tuple(5, NULL);
        tuple->items[0] = MP_OBJ_NEW_SMALL_INT(info->ai_family);
        tuple->items[1] = MP_OBJ_NEW_SMALL_INT(SOCK_STREAM);
        tuple->items[2] = MP_OBJ_NEW_SMALL_INT(IPPROTO_TCP);
        tuple->items[3] = MP_OBJ_NEW_QSTR(MP_QSTR_);
        tuple->items[4] = format_inet_addr(&info->ai_addr, state->port);

        mp_obj_list_append(state->result, MP_OBJ_FROM_PTR(tuple));
        // Do NOT give semaphore here, wait for final info == NULL call
        return;
    }

    // If info is NULL, it signals the end of results for this query type or an error
    if (info == NULL) {
         if (status != DNS_EAI_ALLDONE && state->status == 0) {
            // If it's an error status and we haven't recorded one yet
            state->status = status;
            printk("Setting final error status: %d\n", status);
        } else if (status == DNS_EAI_ALLDONE && state->status == 0 && mp_obj_len(state->result) == 0) {
             // Handle the case where ALLDONE is received but no addresses were found, and no error was set yet.
             // This might indicate a resolution failure even if status was 0 initially.
             state->status = -ENOENT; // Or another appropriate error code for no address found
             printk("Setting status to ENOENT: %d\n", state->status);
        }

        // Always give the semaphore when info is NULL to signal completion for this query type
        printk("Giving semaphore on info == NULL\n");
        k_sem_give(&state->sem);
    }
 }
 
 STATIC mp_obj_t mod_getaddrinfo(size_t n_args, const mp_obj_t *args) {
    mp_obj_t host_in = args[0], port_in = args[1];
    const char *host = mp_obj_str_get_str(host_in);
    mp_int_t family = AF_INET;
    if (n_args > 2) {
        family = mp_obj_get_int(args[2]);
    }

    printk("Using system DNS resolver...\n");
    
    getaddrinfo_state_t state;
    state.port = port_in;
    state.result = mp_obj_new_list(0, NULL);
    state.calling_thread_state = mp_thread_get_state();
    k_sem_init(&state.sem, 0, UINT_MAX);

    for (int i = 2; i--;) {
        int type = (family != AF_INET6 ? DNS_QUERY_TYPE_A : DNS_QUERY_TYPE_AAAA);
        printk("Querying DNS for %s (type %d)...\n", host, type);
        RAISE_ERRNO(dns_get_addr_info(host, type, NULL, dns_resolve_cb, &state, 5000));
        
        printk("semaphore wait...1\n");
        if (k_sem_take(&state.sem, K_SECONDS(10)) != 0) {
            printk("semaphore wait...2\n");
            mp_raise_OSError(ETIMEDOUT);
        }

        if (family != 0) {
            break;
        }
        family = AF_INET6;
    }

    printk("something...1\n");
    mp_int_t len = MP_OBJ_SMALL_INT_VALUE(mp_obj_len(state.result));
    if (state.status != 0 && len == 0) {
        mp_raise_OSError(state.status);
    }

    return state.result;
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_VAR_BETWEEN(mod_getaddrinfo_obj, 2, 3, mod_getaddrinfo);
 
 #if !CONFIG_NET_SOCKETS_OFFLOAD
 STATIC mp_obj_t pkt_get_info(void) {
     struct k_mem_slab *rx, *tx;
     struct net_buf_pool *rx_data, *tx_data;
     net_pkt_get_info(&rx, &tx, &rx_data, &tx_data);
     mp_obj_tuple_t *t = MP_OBJ_TO_PTR(mp_obj_new_tuple(4, NULL));
     t->items[0] = MP_OBJ_NEW_SMALL_INT(k_mem_slab_num_free_get(rx));
     t->items[1] = MP_OBJ_NEW_SMALL_INT(k_mem_slab_num_free_get(tx));
     t->items[2] = MP_OBJ_NEW_SMALL_INT(rx_data->avail_count);
     t->items[3] = MP_OBJ_NEW_SMALL_INT(tx_data->avail_count);
     return MP_OBJ_FROM_PTR(t);
 }
 STATIC MP_DEFINE_CONST_FUN_OBJ_0(pkt_get_info_obj, pkt_get_info);
 #endif
 
 STATIC const mp_rom_map_elem_t mp_module_socket_globals_table[] = {
     { MP_ROM_QSTR(MP_QSTR___name__), MP_ROM_QSTR(MP_QSTR_socket) },
     // objects
     { MP_ROM_QSTR(MP_QSTR_socket), MP_ROM_PTR(&socket_type) },
     // class constants
     { MP_ROM_QSTR(MP_QSTR_AF_INET), MP_ROM_INT(AF_INET) },
     { MP_ROM_QSTR(MP_QSTR_AF_INET6), MP_ROM_INT(AF_INET6) },
     { MP_ROM_QSTR(MP_QSTR_AF_PACKET), MP_ROM_INT(AF_PACKET) },
 
     { MP_ROM_QSTR(MP_QSTR_SOCK_STREAM), MP_ROM_INT(SOCK_STREAM) },
     { MP_ROM_QSTR(MP_QSTR_SOCK_DGRAM), MP_ROM_INT(SOCK_DGRAM) },
     { MP_ROM_QSTR(MP_QSTR_SOCK_RAW), MP_ROM_INT(SOCK_RAW) },
     #if CONFIG_NET_SOCKETS_OFFLOAD
     { MP_ROM_QSTR(MP_QSTR_IPPROTO_TLS_1_2), MP_ROM_INT(IPPROTO_TLS_1_2) },
     { MP_ROM_QSTR(MP_QSTR_TLS_PEER_VERIFY_NONE), MP_ROM_INT(TLS_PEER_VERIFY_NONE) },
     { MP_ROM_QSTR(MP_QSTR_TLS_PEER_VERIFY_OPTIONAL), MP_ROM_INT(TLS_PEER_VERIFY_OPTIONAL) },
     { MP_ROM_QSTR(MP_QSTR_TLS_PEER_VERIFY_REQUIRED), MP_ROM_INT(TLS_PEER_VERIFY_REQUIRED) },
     #endif
 
     { MP_ROM_QSTR(MP_QSTR_SOL_SOCKET), MP_ROM_INT(1) },
     { MP_ROM_QSTR(MP_QSTR_SO_REUSEADDR), MP_ROM_INT(2) },
 
     { MP_ROM_QSTR(MP_QSTR_getaddrinfo), MP_ROM_PTR(&mod_getaddrinfo_obj) },
     #if !CONFIG_NET_SOCKETS_OFFLOAD
     { MP_ROM_QSTR(MP_QSTR_pkt_get_info), MP_ROM_PTR(&pkt_get_info_obj) },
     #endif
 };
 
 STATIC MP_DEFINE_CONST_DICT(mp_module_socket_globals, mp_module_socket_globals_table);
 
 const mp_obj_module_t mp_module_socket = {
     .base = { &mp_type_module },
     .globals = (mp_obj_dict_t *)&mp_module_socket_globals,
 };
 
 MP_REGISTER_EXTENSIBLE_MODULE(MP_QSTR_socket, mp_module_socket);
 
 #endif // (CONFIG_NET_SOCKETS && !CONFIG_EXCLUDE_PY_SOCKETS)